////////////////////////////////////////////////////////////////////////////////
// checkstyle: Checks Java source code for adherence to a set of rules.
// Copyright (C) 2001-2021 the original author or authors.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
////////////////////////////////////////////////////////////////////////////////

package com.puppycrawl.tools.checkstyle.checks.coding;

import com.puppycrawl.tools.checkstyle.StatelessCheck;
import com.puppycrawl.tools.checkstyle.api.AbstractCheck;
import com.puppycrawl.tools.checkstyle.api.DetailAST;
import com.puppycrawl.tools.checkstyle.api.TokenTypes;

/**
 * <p>
 * Detects inline conditionals. Here is one example of an inline conditional:
 * </p>
 * <pre>
 * String a = getParameter("a");
 * String b = (a==null || a.length()&lt;1) ? null : a.substring(1);
 * </pre>
 * <p>
 * Rationale: Some developers find inline conditionals hard to read, so
 * their employer's coding standards forbid them.
 * </p>
 * <p>
 * To configure the check:
 * </p>
 * <pre>
 * &lt;module name=&quot;AvoidInlineConditionals&quot;/&gt;
 * </pre>
 * <p>Example:</p>
 * <pre>
 * int x = 5;
 * boolean foobar = (x == 5); // OK
 *
 * String text;
 * text = (text == null) ? "" : text; // violation
 *
 * String b;
 * if (a != null &amp;&amp; a.length() &gt;= 1) { // OK
 *   b = a.substring(1);
 * } else {
 *   b = null;
 * }
 *
 * b = (a != null &amp;&amp; a.length() &gt;= 1) ? a.substring(1) : null; // violation
 * </pre>
 * <p>
 * Parent is {@code com.puppycrawl.tools.checkstyle.TreeWalker}
 * </p>
 * <p>
 * Violation Message Keys:
 * </p>
 * <ul>
 * <li>
 * {@code inline.conditional.avoid}
 * </li>
 * </ul>
 *
 * @since 3.1
 */
@StatelessCheck
public class AvoidInlineConditionalsCheck extends AbstractCheck {

    /**
     * A key is pointing to the warning message text in "messages.properties"
     * file.
     */
    public static final String MSG_KEY = "inline.conditional.avoid";

    @Override
    public int[] getDefaultTokens() {
        return getRequiredTokens();
    }

    @Override
    public int[] getRequiredTokens() {
        return new int[] {TokenTypes.QUESTION};
    }

    @Override
    public int[] getAcceptableTokens() {
        return getRequiredTokens();
    }

    @Override
    public void visitToken(DetailAST ast) {
        // the only place a QUESTION token can occur is in inline conditionals
        // so no need to do any further tricks here - pretty trivial Check!

        log(ast, MSG_KEY);
    }

}
